from .base_storage_class import BaseStorageClass
from .storage_class_compressed_memory import StorageClassCompressedMemory
from .storage_class_disk import StorageClassDisk
from .storage_class_memory import StorageClassMemory
